from __future__ import print_function, unicode_literals, absolute_import
from .headers import get_headers
import sys
import io
import gzip
import bz2

try:
    import lzma
except ImportError:
    pass
try:
    import zstandard
except ImportError:
    pass
import struct
from rpmfile import cpiofile
from functools import wraps
from rpmfile.io_extra import _SubFile

pad = lambda fileobj: (4 - (fileobj.tell() % 4)) % 4


class NoLZMAModuleError(NotImplementedError):
    pass


class NoZSTANDARDModuleError(NotImplementedError):
    pass


class NoBytesIOError(NotImplementedError):
    pass


class RPMInfo(object):
    """
    Informational class which holds the details about an
    archive member given by an RPM entry block.
    RPMInfo objects are returned by RPMFile.getmember() and
    RPMFile.getmembers() and are
    usually created internally.
    """

    _new_coder = struct.Struct(b"8s8s8s8s8s8s8s8s8s8s8s8s8s")

    def __init__(self, name, file_start, file_size, initial_offset, isdir):
        self.name = name
        self.file_start = file_start
        self.size = file_size
        self.initial_offset = initial_offset
        self._isdir = isdir

    @property
    def isdir(self):
        return self._isdir

    def __repr__(self):
        return "<RPMMember %r>" % self.name

    @classmethod
    def _read(cls, magic, fileobj):
        if magic == b"070701":
            return cls._read_new(fileobj, magic=magic)
        else:
            raise Exception("bad magic number %r" % magic)

    @classmethod
    def _read_new(cls, fileobj, magic=None):
        coder = cls._new_coder

        initial_offset = fileobj.tell()
        d = coder.unpack_from(fileobj.read(coder.size))

        namesize = int(d[11], 16)
        name = fileobj.read(namesize)[:-1].decode("utf-8")
        fileobj.seek(pad(fileobj), 1)
        file_start = fileobj.tell()
        file_size = int(d[6], 16)
        fileobj.seek(file_size, 1)
        fileobj.seek(pad(fileobj), 1)
        # https://www.mankier.com/5/cpio under Old Binary Format mode bits
        mode = int(d[1], 16)
        isdir = mode & int("0040000", 8)
        return cls(name, file_start, file_size, initial_offset, isdir)


class RPMFile(object):
    """
    Open an RPM archive `name'. `mode' must be 'r' to
    read from an existing archive.

    If `fileobj' is given, it is used for reading or writing data. If it
    can be determined, `mode' is overridden by `fileobj's mode.
    `fileobj' is not closed, when TarFile is closed.

    """

    def __init__(self, name=None, mode="rb", fileobj=None):
        if mode != "rb":
            raise NotImplementedError("currently the only supported mode is 'rb'")
        self._fileobj = fileobj or io.open(name, mode)
        self._header_range, self._headers = get_headers(self._fileobj)
        self._ownes_fd = fileobj is None

    @property
    def data_offset(self):
        return self._header_range[1]

    @property
    def header_range(self):
        return self._header_range

    @property
    def headers(self):
        "RPM headers"
        return self._headers

    def __enter__(self):
        return self

    def __exit__(self, *excinfo):
        if self._ownes_fd:
            self._fileobj.close()

    _members = None

    def getmembers(self):
        """
        Return the members of the archive as a list of RPMInfo objects. The
        list has the same order as the members in the archive.
        """
        if self._members is None:
            self._members = _members = []
            g = self.data_file
            magic = g.read(2)
            while magic:
                if magic == b"07":
                    magic += g.read(4)
                    member = RPMInfo._read(magic, g)

                    if member.name == "TRAILER!!!":
                        break

                    if not member.isdir:
                        _members.append(member)

                magic = g.read(2)
            return _members
        return self._members

    def getmember(self, name):
        """
        Return an RPMInfo object for member `name'. If `name' can not be
        found in the archive, KeyError is raised. If a member occurs more
        than once in the archive, its last occurrence is assumed to be the
        most up-to-date version.
        """
        members = self.getmembers()
        for m in members[::-1]:
            if m.name == name:
                return m

        raise KeyError("member %s could not be found" % name)

    def extractfile(self, member):
        """
        Extract a member from the archive as a file object. `member' may be
        a filename or an RPMInfo object.
        The file-like object is read-only and provides the following
        methods: read(), readline(), readlines(), seek() and tell()
        """
        if not isinstance(member, RPMInfo):
            member = self.getmember(member)
        return _SubFile(self.data_file, member.file_start, member.size)

    _data_file = None

    @property
    def data_file(self):
        """Return the uncompressed raw CPIO data of the RPM archive."""

        if self._data_file is None:
            fileobj = _SubFile(self._fileobj, self.data_offset)

            if self.headers["archive_compression"] == b"xz":
                if not getattr(sys.modules[__name__], "lzma", False):
                    raise NoLZMAModuleError("lzma module not present")
                self._data_file = lzma.LZMAFile(fileobj)
            elif self.headers["archive_compression"] == b"zstd":
                if not getattr(sys.modules[__name__], "zstandard", False):
                    raise NoZSTANDARDModuleError("zstandard module not present")
                if not (sys.version_info.major >= 3 and sys.version_info.minor >= 5):
                    raise NoBytesIOError("Need io.BytesIO (Python >= 3.5)")
                with zstandard.ZstdDecompressor().stream_reader(fileobj) as zstd_data:
                    self._data_file = io.BytesIO(zstd_data.read())

            elif self.headers["archive_compression"] == b"bzip2":
                self._data_file = bz2.BZ2File(fileobj)
            else:
                self._data_file = gzip.GzipFile(fileobj=fileobj)

        return self._data_file


def open(name=None, mode="rb", fileobj=None):
    """
    Open an RPM archive for reading. Return
    an appropriate RPMFile class.
    """
    return RPMFile(name, mode, fileobj)


def main():
    print(sys.argv[1])
    with open(sys.argv[1]) as rpm:
        print(rpm.headers)
        for m in rpm.getmembers():
            print(m)
        print("done")


if __name__ == "__main__":
    main()



"""
Created on Jan 10, 2014

@author: sean
"""
from __future__ import print_function, unicode_literals, absolute_import
import sys
import struct
from pprint import pprint
from .errors import RPMError

tags = {
    "headerimage": 61,
    "headersignatures": 62,
    "headerimmutable": 63,
    "headerregions": 64,
    "headeri18ntable": 100,
    "sig_base": 256,
    "sigsize": 257,
    "sigpgp": 259,
    "siggpg": 262,
    "pubkeys": 266,
    "signature": 267,
    "rsaheader": 268,
    "md5": 269,
    "longsigsize": 270,
    "longarchivesize": 271,
    "name": 1000,
    "version": 1001,
    "release": 1002,
    "serial": 1003,
    "summary": 1004,
    "description": 1005,
    "buildtime": 1006,
    "buildhost": 1007,
    "installtime": 1008,
    "size": 1009,
    "distribution": 1010,
    "vendor": 1011,
    "gif": 1012,
    "xpm": 1013,
    "copyright": 1014,
    "packager": 1015,
    "group": 1016,
    "changelog": 1017,
    "source": 1018,
    "patch": 1019,
    "url": 1020,
    "os": 1021,
    "arch": 1022,
    "prein": 1023,
    "postin": 1024,
    "preun": 1025,
    "postun": 1026,
    "filenames": 1027,
    "filesizes": 1028,
    "filestates": 1029,
    "filemodes": 1030,
    "fileuids": 1031,
    "filegids": 1032,
    "filerdevs": 1033,
    "filemtimes": 1034,
    "filemd5s": 1035,
    "filelinktos": 1036,
    "fileflags": 1037,
    "root": 1038,
    "fileusername": 1039,
    "filegroupname": 1040,
    "exclude": 1041,
    "exclusive": 1042,
    "icon": 1043,
    "sourcerpm": 1044,
    "fileverifyflags": 1045,
    "archivesize": 1046,
    "provides": 1047,
    "requireflags": 1048,
    "requirename": 1049,
    "requireversion": 1050,
    "nosource": 1051,
    "nopatch": 1052,
    "conflictflags": 1053,
    "conflictname": 1054,
    "conflictversion": 1055,
    "defaultprefix": 1056,
    "buildroot": 1057,
    "installprefix": 1058,
    "excludearch": 1059,
    "excludeos": 1060,
    "exclusivearch": 1061,
    "exclusiveos": 1062,
    "autoreqprov": 1063,
    "rpmversion": 1064,
    "triggerscripts": 1065,
    "triggername": 1066,
    "triggerversion": 1067,
    "triggerflags": 1068,
    "triggerindex": 1069,
    "verifyscript": 1079,
    "changelogtime": 1080,
    "authors": 1081,
    "comments": 1082,
    "prereq": 1084,
    "preinprog": 1085,
    "postinprog": 1086,
    "preunprog": 1087,
    "postunprog": 1088,
    "buildarchs": 1089,
    "obsoletes": 1090,
    "verifyscriptprog": 1091,
    "triggerscriptprog": 1092,
    "docdir": 1093,
    "cookie": 1094,
    "filedevices": 1095,
    "fileinodes": 1096,
    "filelangs": 1097,
    "prefixes": 1098,
    "instprefixes": 1099,
    "triggerin": 1100,
    "triggerun": 1101,
    "triggerpostun": 1102,
    "autoreq": 1103,
    "autoprov": 1104,
    "capability": 1105,
    "sourcepackage": 1106,
    "oldorigfilenames": 1107,
    "buildprereq": 1108,
    "buildrequires": 1109,
    "buildconflicts": 1110,
    "buildmacros": 1111,
    "provideflags": 1112,
    "provideversion": 1113,
    "obsoleteflags": 1114,
    "obsoleteversion": 1115,
    "dirindexes": 1116,
    "basenames": 1117,
    "dirnames": 1118,
    "origdirindexes": 1119,
    "origbasenames": 1120,
    "origdirnames": 1121,
    "optflags": 1122,
    "disturl": 1123,
    "archive_format": 1124,
    "archive_compression": 1125,
    "payloadflags": 1126,
    "installcolor": 1127,
    "installtid": 1128,
    "removetid": 1129,
    "rhnplatform": 1131,
    "target": 1132,
    "patchesname": 1133,
    "patchesflags": 1134,
    "patchesversion": 1135,
    "cachectime": 1136,
    "cachepkgpath": 1137,
    "cachepkgsize": 1138,
    "cachepkgmtime": 1139,
    "filecolors": 1140,
    "fileclass": 1141,
    "classdict": 1142,
    "filedependsx": 1143,
    "filedependsn": 1144,
    "dependsdict": 1145,
    "sourcepkgid": 1146,
    "filecontexts": 1147,
    "fscontexts": 1148,
    "recontexts": 1149,
    "policies": 1150,
    "pretrans": 1151,
    "posttrans": 1152,
    "pretransprog": 1153,
    "posttransprog": 1154,
    "disttag": 1155,
    "oldsuggestsname": 1156,
    "oldsuggestsversion": 1157,
    "oldsuggestsflags": 1158,
    "oldenhancesname": 1159,
    "oldenhancesversion": 1160,
    "oldenhancesflags": 1161,
    "priority": 1162,
    "cvsid": 1163,
    "blinkpkgid": 1164,
    "blinkhdrid": 1165,
    "blinknevra": 1166,
    "flinkpkgid": 1167,
    "flinkhdrid": 1168,
    "flinknevra": 1169,
    "packageorigin": 1170,
    "triggerprein": 1171,
    "buildsuggests": 1172,
    "buildenhances": 1173,
    "scriptstates": 1174,
    "scriptmetrics": 1175,
    "buildcpuclock": 1176,
    "filedigestalgos": 1177,
    "variants": 1178,
    "xmajor": 1179,
    "xminor": 1180,
    "repotag": 1181,
    "keywords": 1182,
    "buildplatforms": 1183,
    "packagecolor": 1184,
    "packageprefcolor": 1185,
    "xattrsdict": 1186,
    "filexattrsx": 1187,
    "depattrsdict": 1188,
    "conflictattrsx": 1189,
    "obsoleteattrsx": 1190,
    "provideattrsx": 1191,
    "requireattrsx": 1192,
    "buildprovides": 1193,
    "buildobsoletes": 1194,
    "dbinstance": 1195,
    "nvra": 1196,
    "filenames": 5000,
    "fileprovide": 5001,
    "filerequire": 5002,
    "fsnames": 5003,
    "fssizes": 5004,
    "triggerconds": 5005,
    "triggertype": 5006,
    "origfilenames": 5007,
    "longfilesizes": 5008,
    "longsize": 5009,
    "filecaps": 5010,
    "filedigestalgo": 5011,
    "bugurl": 5012,
    "evr": 5013,
    "nvr": 5014,
    "nevr": 5015,
    "nevra": 5016,
    "headercolor": 5017,
    "verbose": 5018,
    "epochnum": 5019,
    "preinflags": 5020,
    "postinflags": 5021,
    "preunflags": 5022,
    "postunflags": 5023,
    "pretransflags": 5024,
    "posttransflags": 5025,
    "verifyscriptflags": 5026,
    "triggerscriptflags": 5027,
    "collections": 5029,
    "policynames": 5030,
    "policytypes": 5031,
    "policytypesindexes": 5032,
    "policyflags": 5033,
    "vcs": 5034,
    "ordername": 5035,
    "orderversion": 5036,
    "orderflags": 5037,
    "mssfmanifest": 5038,
    "mssfdomain": 5039,
    "instfilenames": 5040,
    "requirenevrs": 5041,
    "providenevrs": 5042,
    "obsoletenevrs": 5043,
    "conflictnevrs": 5044,
    "filenlinks": 5045,
    "recommendname": 5046,
    "recommendversion": 5047,
    "recommendflags": 5048,
    "suggestname": 5049,
    "suggestversion": 5050,
    "suggestflags": 5051,
    "supplementname": 5052,
    "supplementversion": 5053,
    "supplementflags": 5054,
    "enhancename": 5055,
    "enhanceversion": 5056,
    "enhanceflags": 5057,
    "recommendnevrs": 5058,
    "suggestnevrs": 5059,
    "supplementnevrs": 5060,
    "enhancenevrs": 5061,
    "encoding": 5062,
    "filetriggerin": 5063,
    "filetriggerun": 5064,
    "filetriggerpostun": 5065,
    "filetriggerscripts": 5066,
    "filetriggerscriptprog": 5067,
    "filetriggerscriptflags": 5068,
    "filetriggername": 5069,
    "filetriggerindex": 5070,
    "filetriggerversion": 5071,
    "filetriggerflags": 5072,
    "transfiletriggerin": 5073,
    "transfiletriggerun": 5074,
    "transfiletriggerpostun": 5075,
    "transfiletriggerscripts": 5076,
    "transfiletriggerscriptprog": 5077,
    "transfiletriggerscriptflags": 5078,
    "transfiletriggername": 5079,
    "transfiletriggerindex": 5080,
    "transfiletriggerversion": 5081,
    "transfiletriggerflags": 5082,
    "removepathpostfixes": 5083,
    "filetriggerpriorities": 5084,
    "transfiletriggerpriorities": 5085,
    "filetriggerconds": 5086,
    "filetriggertype": 5087,
    "transfiletriggerconds": 5088,
    "transfiletriggertype": 5089,
    "filesignatures": 5090,
    "filesignaturelength": 5091,
    "payloaddigest": 5092,
    "payloaddigestalgo": 5093,
    "autoinstalled": 5094,
    "identity": 5095,
    "modularitylabel": 5096,
}

rtags = dict([(value, key) for (key, value) in tags.items()])


def extract_string(offset, count, store):
    if count > 1:
        return extract_array(offset, count, store)
    assert count == 1
    idx = store[offset:].index(b"\x00")
    return store[offset : offset + idx]


def extract_i18nstring(offset, count, store):
    # rpm string header entries can have multiple versions, one for each locale.
    # the locale names are defined in the i18n table header entry. For the sake of
    # simplicity, take only one locale to use
    return store[offset:].split(b"\x00", count)[0]


def extract_array(offset, count, store):
    return store[offset:].split(b"\x00", count)[:-1]


def extract_bin(offset, count, store):
    return store[offset : offset + count]


def extract_int32(offset, count, store):
    values = struct.unpack(b"!" + b"i" * count, store[offset : offset + 4 * count])
    if count == 1:
        values = values[0]
    return values


def extract_int16(offset, count, store):
    values = struct.unpack(b"!" + b"h" * count, store[offset : offset + 2 * count])
    if count == 1:
        values = values[0]
    return values


ty_map = {
    3: extract_int16,
    4: extract_int32,
    6: extract_string,
    7: extract_bin,
    8: extract_array,
    9: extract_i18nstring,
}


def extract_data(ty, offset, count, store):
    extract = ty_map.get(ty)
    if extract:
        return extract(offset, count, store)
    else:
        return "could not extract %s" % ty


def _readheader(fileobj):
    char = fileobj.read(1)
    while char != b"\x8e":
        char = fileobj.read(1)

        if char is None or char == b"":
            raise RPMError("reached end of file without finding magic char \x8e")

    magic = b"\x8e" + fileobj.read(2)
    from binascii import hexlify

    assert hexlify(magic) == b"8eade8", hexlify(magic)
    version = ord(fileobj.read(1))

    header_start = fileobj.tell() - 4  # -4 for magic

    _ = fileobj.read(4)

    (num_entries,) = struct.unpack(b"!i", fileobj.read(4))
    (header_structure_size,) = struct.unpack(b"!i", fileobj.read(4))

    header = struct.Struct(b"!iiii")

    entries = []
    for _ in range(num_entries):
        entry = header.unpack(fileobj.read(header.size))
        entries.append(entry)

    store = fileobj.read(header_structure_size)
    store

    headers = {}
    for tag, ty, offset, count in entries:
        key = rtags.get(tag, tag)
        value = extract_data(ty, offset, count, store)
        headers[key] = value
    header_end = fileobj.tell()
    return (header_start, header_end), headers


def get_headers(fileobj):
    lead = struct.Struct(b"!4sBBhh66shh16s")
    data = fileobj.read(lead.size)
    value = lead.unpack(data)

    # Not sure what the first set of headers are for
    first_range, first_headers = _readheader(fileobj)
    second_range, second_headers = _readheader(fileobj)

    first_headers.update(second_headers)

    return second_range, first_headers


def main():
    with open(sys.argv[1]) as fileobj:
        headers = get_headers(fileobj)
        pprint(headers)


if __name__ == "__main__":
    main()



#!/usr/bin/env python -3
# -*- coding: utf-8 -*-
#
# Copyright Â© 2011, 2013 K Richard Pixley
#
# See LICENSE for details.
#
# Time-stamp: <30-Jun-2013 19:07:22 PDT by rich@noir.com>

"""
Cpiofile is a library which reads and writes unix style 'cpio' format
archives.

.. todo:: open vs context manager
.. todo:: make is_cpiofile work on fileobj
"""

from __future__ import unicode_literals, print_function

__docformat__ = "restructuredtext en"

__all__ = [
    "CheckSumError",
    "CpioError",
    "CpioFile",
    "CpioMember",
    "HeaderError",
    "InvalidFileFormat",
    "InvalidFileFormatNull",
    "is_cpiofile",
    "valid_magic",
]

import abc
import io
import mmap
import os
import struct


class CpioError(Exception):
    """Base class for CpioFile exceptions"""

    pass


class CheckSumError(CpioError):
    """Exception indicating a check sum error"""

    pass


class InvalidFileFormat(CpioError):
    """Exception indicating a file format error"""

    pass


class InvalidFileFormatNull(InvalidFileFormat):
    """Exception indicating a null file"""

    pass


class HeaderError(CpioError):
    """Exception indicating a header error"""

    pass


def valid_magic(block):
    """predicate indicating whether *block* includes a valid magic number"""
    return CpioMember.valid_magic(block)


def is_cpiofile(name):
    """predicate indicating whether *name* is a valid cpiofile"""
    with io.open(name, "rb") as fff:
        return valid_magic(fff.read(16))


class StructBase(object):
    """
    An abstract base class representing objects which are inherently
    based on a struct.
    """

    __metaclass__ = abc.ABCMeta

    coder = None
    """
    The :py:class:`struct.Struct` used to encode/decode this object
    into a block of memory.  This is expected to be overridden by
    subclasses.
    """  # pylint: disable=W0105

    @property
    def size(self):
        """
        Exact size in bytes of a block of memory into which is suitable
        for packing this instance.
        """
        return self.coder.size

    def unpack(self, block):
        """convenience function for unpacking"""
        return self.unpack_from(block)

    @abc.abstractmethod
    def unpack_from(self, block, offset=0):
        """
        Set the values of this instance from an in-memory
        representation of the struct.

        :param string block: block of memory from which to unpack
        :param int offset: optional offset into the memory block from
            which to start unpacking
        """
        raise NotImplementedError

    def pack(self):
        """convenience function for packing"""
        block = bytearray(self.size)
        self.pack_into(block)
        return block

    @abc.abstractmethod
    def pack_into(self, block, offset=0):
        """
        Store the values of this instance into an in-memory
        representation of the file.

        :param string block: block of memory into which to pack
        :param int offset: optional offset into the memory block into
            which to start packing
        """
        raise NotImplementedError

    __hash__ = None

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)

    def close_enough(self, other):
        """
        This is a comparison similar to __eq__ except that here the
        goal is to determine whether two objects are "close enough"
        despite perhaps having been produced at different times in
        different locations in the file system.
        """
        return self == other


class CpioFile(StructBase):
    """Class representing an entire cpio file"""

    _members = []

    def __init__(self):
        self._members = []

    @property
    def members(self):
        """accessor for a list of the members of this cpio file"""
        return self._members

    @property
    def names(self):
        """accessor for a list of names of the members of this cpio file"""
        return [member.name for member in self.members]

    def __enter__(self):
        return self

    def __exit__(self, thingy, value, traceback):
        self.close()

    @classmethod
    def open(cls, name=None, mode=None):
        return cls._open(cls(), name)

    def _open(self, name=None, fileobj=None, mymap=None, block=None):
        """
        The _open function takes some form of file identifier and creates
        an :py:class:`CpioFile` instance from it.

        :param :py:class:`str` name: a file name
        :param :py:class:`file` fileobj: if given, this overrides *name*
        :param :py:class:`mmap.mmap` mymap: if given, this overrides *fileobj*
        :param :py:class:`bytes` block: file contents in a block of memory, (if given, this overrides *mymap*)

        The file to be used can be specified in any of four different
        forms, (in reverse precedence):

        #. a file name
        #. :py:class:`file` object
        #. :py:mod:`mmap.mmap`, or
        #. a block of memory
        """

        if block is not None:
            if not name:
                name = "<unknown>"

            self.unpack_from(block)

            if fileobj:
                fileobj.close()

            return self

        if mymap is not None:
            block = mymap

        elif fileobj:
            try:
                mymap = mmap.mmap(fileobj.fileno(), 0, mmap.MAP_SHARED, mmap.PROT_READ)

            # pylint: disable=W0702
            except:
                mymap = 0
                block = fileobj.read()

        elif name:
            fileobj = io.open(os.path.normpath(os.path.expanduser(name)), "rb")

        else:
            assert False

        return self._open(name=name, fileobj=fileobj, mymap=mymap, block=block)

    def close(self):
        """noop - here for completeness"""
        pass

    def unpack_from(self, block, offset=0):
        pointer = offset
        print("unpack_from")
        while "TRAILER!!!" not in self.names:
            cmem = CpioMember.encoded_class(block, pointer)()
            print(type(cmem))
            self.members.append(cmem.unpack_from(block, pointer))
            pointer += cmem.size

        del self.members[-1]

    def pack_into(self, block, offset=0):
        pointer = offset

        for member in self.members:
            member.pack_into(block, pointer)
            pointer += member.size

        cmtype = type(self.members[0]) if self.members else CpioMemberNew
        cmt = cmtype()
        cmt.name = "TRAILER!!!"
        cmt.pack_into(block, pointer)

    def get_member(self, name):
        """return a member by *name*"""
        for member in self.members:
            if member.name == name:
                return member

        return None

    def __eq__(self, other):
        raise NotImplementedError


class CpioMember(StructBase):
    """class representing a member of a cpio archive"""

    coder = None

    name = None
    magic = None
    devmajor = None
    devminor = None
    ino = None
    mode = None
    uid = None
    gid = None
    nlink = None
    rdevmajor = None
    rdevminor = None
    mtime = None
    filesize = None

    content = None

    @staticmethod
    def valid_magic(block, offset=0):
        """
        predicate indicating whether a block of memory has a valid magic number
        """
        try:
            return CpioMember.encoded_class(block, offset)
        except InvalidFileFormat:
            return False

    @staticmethod
    def encoded_class(block, offset=0):
        """
        predicate indicating whether a block of memory includes a magic number
        """
        if not block:
            raise InvalidFileFormatNull

        for key in __magicmap__:
            if block.find(key, offset, offset + len(key)) > -1:
                return __magicmap__[key]

        raise InvalidFileFormat

    def unpack_from(self, block, offset=0):
        (
            self.magic,
            dev,
            self.ino,
            self.mode,
            self.uid,
            self.gid,
            self.nlink,
            rdev,
            mtimehigh,
            mtimelow,
            namesize,
            filesizehigh,
            filesizelow,
        ) = self.coder.unpack_from(block, offset)

        self.devmajor = os.major(dev)
        self.devminor = os.minor(dev)
        self.rdevmajor = os.major(rdev)
        self.rdevminor = os.minor(rdev)

        self.mtime = (mtimehigh << 16) | mtimelow
        self.filesize = (filesizehigh << 16) | filesizelow

        namestart = offset + self.coder.size
        datastart = namestart + namesize

        self.name = block[namestart : datastart - 1]  # drop the null

        if isinstance(self, CpioMemberBin) and (namesize & 1):
            datastart += 1  # skip a pad byte if necessary

        self.content = block[datastart : datastart + self.filesize]

        return self

    def pack_into(self, block, offset=0):
        namesize = len(self.name)
        dev = os.makedev(self.devmajor, self.devminor)
        rdev = os.makedev(self.rdevmajor, self.rdevminor)

        mtimehigh = self.mtime >> 16
        mtimelow = self.mtime & 0xFFFF

        filesizehigh = self.filesize >> 16
        filesizelow = self.filesize & 0xFFFF

        self.coder.pack_into(
            block,
            offset,
            self.magic,
            dev,
            self.ino,
            self.mode,
            self.uid,
            self.gid,
            self.nlink,
            rdev,
            mtimehigh,
            mtimelow,
            namesize,
            filesizehigh,
            filesizelow,
        )

        namestart = offset + self.coder.size
        datastart = namestart + namesize + 1

        block[namestart : datastart - 1] = self.name
        block[datastart - 1] = "\x00"

        if isinstance(self, CpioMemberBin) and (namesize & 1):
            datastart += 1
            block[datastart - 1] = "\x00"

        block[datastart : datastart + self.filesize] = self.content

        if isinstance(self, CpioMemberBin) and (self.filesize & 1):
            block[datastart + self.filesize] = "\x00"

        return self

    @property
    def size(self):
        return self.coder.size + len(self.name) + 1 + self.filesize

    def __repr__(self):
        return (
            b"<{0}@{1}: coder={2}, name='{3}', magic='{4}'"
            + ", devmajor={5}, devminor={6}, ino={7}, mode={8}"
            + ", uid={9}, gid={10}, nlink={11}, rdevmajor={12}"
            + ", rdevmino={13}, mtime={14}, filesize={15}>".format(
                self.__class__.__name__,
                hex(id(self)),
                self.coder,
                self.name,
                self.magic,
                self.devmajor,
                self.devminor,
                self.ino,
                self.mode,
                self.uid,
                self.gid,
                self.nlink,
                self.rdevmajor,
                self.rdevminor,
                self.mtime,
                self.filesize,
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.coder == other.coder
            and self.magic == other.magic
            and self.devmajor == other.devmajor
            and self.devminor == other.devminor
            and self.ino == other.ino
            and self.mode == other.mode
            and self.uid == other.uid
            and self.gid == other.gid
            and self.nlink == other.nlink
            and self.rdevmajor == other.rdevmajor
            and self.rdevminor == other.rdevminor
            and self.mtime == other.mtime
            and self.filesize == other.filesize
        )

    close_enough = __eq__


class CpioMemberBin(CpioMember):
    """intermediate class indicating binary members - for subclassing only"""

    @property
    def size(self):
        namesize = len(self.name) + 1  # add null

        retval = self.coder.size
        retval += namesize

        if isinstance(self, CpioMemberBin) and (namesize & 1):
            retval += 1

        retval += self.filesize

        if isinstance(self, CpioMemberBin) and (self.filesize & 1):
            retval += 1

        return retval


class CpioMember32b(CpioMemberBin):
    """
    .. todo:: need to pad after name and after content for old binary.
    """

    coder = struct.Struct(b">2sHHHHHHHHHHHH")


class CpioMember32l(CpioMemberBin):
    """class representing a 32bit little endian binary member"""

    coder = struct.Struct(b"<2sHHHHHHHHHHHH")


class CpioMemberODC(CpioMember):
    """class representing an ODC member"""

    coder = struct.Struct(b"=6s6s6s6s6s6s6s6s11s6s11s")

    def unpack_from(self, block, offset=0):
        (
            self.magic,
            dev,
            ino,
            mode,
            uid,
            gid,
            nlink,
            rdev,
            mtime,
            namesize,
            filesize,
        ) = self.coder.unpack_from(block, offset)
        _namesize = namesize
        self.ino = int(ino, 8)
        self.mode = int(mode, 8)
        self.uid = int(uid, 8)
        self.gid = int(gid, 8)
        self.nlink = int(nlink, 8)

        dev = int(dev, 8)
        rdev = int(rdev, 8)
        self.devmajor = os.major(dev)
        self.devminor = os.minor(dev)
        self.rdevmajor = os.major(rdev)
        self.rdevminor = os.minor(rdev)

        self.mtime = int(mtime, 8)
        namesize = int(namesize, 8)
        self.filesize = int(filesize, 8)

        namestart = offset + self.coder.size
        datastart = namestart + namesize

        self.name = block[namestart : datastart - 1]  # drop the null
        print("+", _namesize, self.name)
        self.content = block[datastart : datastart + self.filesize]

        return self

    def pack_into(self, block, offset=0):
        dev = os.makedev(self.devmajor, self.devminor)
        ino = str(self.ino)
        mode = str(self.mode)
        uid = str(self.uid)
        gid = str(self.gid)
        nlink = str(self.nlink)
        rdev = os.makedev(self.rdevmajor, self.rdevminor)
        mtime = str(self.mtime)
        namesize = str(len(self.name) + 1)  # add a null
        filesize = str(self.filesize)

        self.coder.pack_into(
            block,
            offset,
            self.magic,
            dev,
            ino,
            mode,
            uid,
            gid,
            nlink,
            rdev,
            mtime,
            namesize,
            filesize,
        )

        namesize = len(self.name) + 1

        namestart = offset + self.coder.size
        datastart = namestart + namesize

        block[namestart : datastart - 2] = self.name
        block[datastart - 1] = "\x00"
        block[datastart : datastart + self.filesize] = self.content

        return self


class CpioMemberNew(CpioMember):
    """class representing a new member"""

    coder = struct.Struct(b"6s8s8s8s8s8s8s8s8s8s8s8s8s8s")

    # pylint: disable=W0613
    @staticmethod
    def _checksum(block, offset, length):
        """return a checksum for *block* at *offset* and *length*"""
        return 0

    # pylint: enable=W0613

    def unpack_from(self, block, offset=0):
        unpacks = self.coder.unpack_from(block, offset)

        self.magic = unpacks[0]

        self.ino = int(unpacks[1], 16)
        self.mode = int(unpacks[2], 16)
        self.uid = int(unpacks[3], 16)
        self.gid = int(unpacks[4], 16)
        self.nlink = int(unpacks[5], 16)

        self.mtime = int(unpacks[6], 16)
        self.filesize = int(unpacks[7], 16)

        self.devmajor = int(unpacks[8], 16)
        self.devminor = int(unpacks[9], 16)
        self.rdevmajor = int(unpacks[10], 16)
        self.rdevminor = int(unpacks[11], 16)

        namesize = int(unpacks[12], 16)
        check = int(unpacks[13], 16)

        namestart = offset + self.coder.size
        nameend = namestart + namesize
        datastart = nameend + ((4 - (nameend % 4)) % 4)  # pad
        dataend = datastart + self.filesize

        self.name = block[namestart : nameend - 1]  # drop the null
        print("name", namesize, self.name)
        print("pad", ((4 - (nameend % 4)) % 4))  # pad
        self.content = block[datastart:dataend]

        if check != self._checksum(self.content, 0, self.filesize):
            raise CheckSumError

        return self

    def pack_into(self, block, offset=0):
        namesize = len(self.name) + 1
        # unused: rdev = os.makedev(self.rdevmajor, self.rdevminor)
        self.coder.pack_into(
            block,
            offset,
            self.magic,
            str(self.ino),
            str(self.mode),
            str(self.uid),
            str(self.gid),
            str(self.nlink),
            str(self.mtime),
            str(self.filesize),
            str(self.devmajor),
            str(self.devminor),
            str(self.rdevmajor),
            str(self.rdevminor),
            str(namesize),
            self._checksum(self.content, 0, self.filesize),
        )

        namestart = offset + self.coder.size
        nameend = namestart + namesize
        datastart = nameend + ((4 - (nameend % 4)) % 4)  # pad
        dataend = datastart + self.filesize

        block[namestart:nameend] = self.name

        for i in range(nameend, datastart):
            block[i] = "\x00"

        block[datastart:dataend] = self.content

        padend = dataend + ((4 - (datastart % 4)) % 4)  # pad
        for i in range(dataend, padend):
            block[i] = "\x00"

        return self

    @property
    def size(self):
        retval = self.coder.size
        retval += len(self.name) + 1
        retval += (4 - (retval % 4)) % 4
        retval += self.filesize
        retval += (4 - (retval % 4)) % 4
        return retval


class CpioMemberCRC(CpioMemberNew):
    """class representing a cpio archive member with a CRC"""

    @staticmethod
    def _checksum(block, offset, length):
        csum = 0

        for i in range(length):
            csum += ord(block[offset + i])

        return csum & 0xFFFFFFFF


__magicmap__ = {
    b"\x71\xc7": CpioMember32b,
    b"\xc7\x71": CpioMember32l,
    b"070707": CpioMemberODC,
    b"070701": CpioMemberNew,
    b"070702": CpioMemberCRC,
}



"""
Created on Jan 11, 2014

@author: sean
"""
import unittest

import rpmfile
import io


class Test(unittest.TestCase):
    def test_seek(self):
        fd = io.BytesIO(b"Hello world")
        sub = rpmfile._SubFile(fd, start=2, size=4)

        sub.seek(0)
        self.assertEqual(sub.tell(), 0)

        sub.seek(1)
        self.assertEqual(sub.tell(), 1)

        sub.seek(1, 1)
        self.assertEqual(sub.tell(), 2)

        sub.seek(-1, 1)
        self.assertEqual(sub.tell(), 1)

        sub.seek(-10, 1)
        self.assertEqual(sub.tell(), 0)

    def test_read(self):
        fd = io.BytesIO(b"Hello world")
        sub = rpmfile._SubFile(fd, start=2, size=4)

        self.assertEqual(sub.read(), b"llo ")
        self.assertEqual(sub.read(), b"")

        sub.seek(0)
        self.assertEqual(sub.read(2), b"ll")

        sub.seek(0)
        self.assertEqual(sub.read(10), b"llo ")


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testSeek']
    unittest.main()



"""
Created on Jan 11, 2014

@author: sean
"""
import io


def _doc(from_func):
    """copy doc from one function to another
    use as a decorator eg::

        @_doc(file.tell)
        def tell(..):
            ...
    """

    def decorator(to_func):
        to_func.__doc__ = from_func.__doc__
        return to_func

    return decorator


class _SubFile(object):
    """A thin wrapper around an existing file object that
    provides a part of its data as an individual file
    object.
    """

    def __init__(self, fileobj, start=0, size=None):
        self._fileobj = fileobj
        self._start = start
        if size is None:
            fileobj.seek(0, 2)
            pos = fileobj.tell()
            self._size = pos - start
        else:
            self._size = size
        self._pos = 0

    def __enter__(self):
        return self

    def __exit__(self, *excinfo):
        pass

    def close(self):
        pass

    def __getattr__(self, attr):
        return getattr(self._fileobj, attr)

    @_doc(io.FileIO.tell)
    def tell(self):
        return self._pos

    @_doc(io.FileIO.seek)
    def seek(self, offset, whence=0):
        if whence == 0:
            self._pos = offset
        elif whence == 1:
            self._pos += offset
        else:
            self._pos = self._size + offset

        self._pos = max(0, self._pos)

    def _n(self, size=None):
        if not size:
            size = self._size
        return min(size, self._size - self._pos)

    @_doc(io.FileIO.read)
    def read(self, size=None):
        self._fileobj.seek(self._pos + self._start, 0)

        n = self._n(size)
        self._pos += n

        return self._fileobj.read(n)

    @_doc(io.FileIO.readline)
    def readline(self, size=None):
        self._fileobj.seek(self._pos + self._start, 0)
        n = self._n(size)
        line = self._fileobj.readline(n)
        self._pos += len(line)
        return line

    @_doc(io.FileIO.readlines)
    def readlines(self, size=None):
        n = self._n(size)
        line = self.readline(n)
        n -= len(line)
        while line:
            yield line
            line = self.readline(n)
            n -= len(line)



import os
import shutil
import tempfile
import time
import unittest

from rpmfile.cli import main

from .test_extract import download


class TempCLI(unittest.TestCase):
    GOPACKET_LICENSE_DIRS = [".", "usr", "share", "doc", "gopacket-license"]
    GOPACKET_LICENSE_FILES = ["AUTHORS", "LICENSE"]
    GOPACKET_LICENSE_INFO = (
        "Name        : gopacket-license\n"
        "Version     : 2019_04_08T07_36_42Z\n"
        "Release     : 1\n"
        "Architecture: noarch\n"
        "Group       : default\n"
        "Size        : 3223\n"
        "License     : BSD\n"
        "Signature   : None\n"
        "Source RPM  : gopacket-license-2019_04_08T07_36_42Z-1.src.rpm\n"
        "Build Date  : Tue Apr  9 08:55:16 2019\n"
        "Build Host  : jenkins-slave-fat-cloud-nlbzt\n"
        "URL         : http://example.com/no-uri-given\n"
        "Summary     : License for gopacket-license\n"
        "Description : \n"
        "License for gopacket-license\n"
    )

    def setUp(cls):
        cls.prevdir = os.getcwd()
        cls.tempdir = tempfile.mkdtemp()
        os.chdir(cls.tempdir)
        os.environ["LC_ALL"] = "C"
        os.environ["TZ"] = "UTC"
        time.tzset()

    def tearDown(cls):
        shutil.rmtree(cls.tempdir)
        os.chdir(cls.prevdir)

    @download(
        "https://github.com/srossross/rpmfile/files/3150016/gopacket-license.noarch.rpm.gz",
        "gopacket.rpm",
    )
    def test_extract(self, rpmpath):
        """That the command line extracts correctly"""
        dest = os.path.join(self.tempdir, "mydir")
        os.mkdir(dest)

        _args, output = main("-xC", dest, rpmpath)

        extracted_files = os.listdir(
            os.path.join(self.tempdir, "mydir", *self.GOPACKET_LICENSE_DIRS)
        )
        self.assertEqual(self.GOPACKET_LICENSE_FILES, sorted(extracted_files))

        self.assertEqual(len(output["extracted"]), len(self.GOPACKET_LICENSE_FILES))
        for filename in self.GOPACKET_LICENSE_FILES:
            self.assertIn(self.GOPACKET_LICENSE_DIRS + [filename], output["extracted"])

    @download(
        "https://github.com/srossross/rpmfile/files/3150016/gopacket-license.noarch.rpm.gz",
        "gopacket.rpm",
    )
    def test_list(self, rpmpath):
        """That the command line lists correctly"""
        _args, output = main("-l", rpmpath)

        self.assertEqual(len(output["list"]), len(self.GOPACKET_LICENSE_FILES))
        for filename in self.GOPACKET_LICENSE_FILES:
            self.assertIn(self.GOPACKET_LICENSE_DIRS + [filename], output["list"])

    @download(
        "https://github.com/srossross/rpmfile/files/3150016/gopacket-license.noarch.rpm.gz",
        "gopacket.rpm",
    )
    def test_info(self, rpmpath):
        """That the command line get RPM infomation correctly"""
        _args, output = main("-i", rpmpath)
        self.assertEqual(output["info"], self.GOPACKET_LICENSE_INFO)



from __future__ import print_function

from datetime import datetime
import os
import io
import sys
import argparse

import rpmfile


def console_script_entry_point():
    main(*sys.argv[1:])


def main(*argv):
    parser = argparse.ArgumentParser(prog="rpmfile")
    parser.add_argument("infile")
    parser.add_argument(
        "-x",
        "--extract",
        dest="extract",
        action="store_true",
        help="Extract the input RPM",
    )
    parser.add_argument(
        "-C",
        "--directory",
        type=str,
        dest="dest",
        help="Extract to this directory when extracting files",
        default=".",
    )
    parser.add_argument(
        "-l",
        "--list",
        dest="list",
        action="store_true",
        help="List files in RPM without extracting",
    )
    parser.add_argument(
        "-i",
        "--info",
        dest="info",
        action="store_true",
        help="Display RPM information without extracting",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Print filenames when extracting",
    )
    args = parser.parse_args(argv)

    if args.infile == "-":
        args.infile = sys.stdin
        if sys.version_info.major >= 3:
            args.infile = args.infile.buffer
    else:
        args.infile = open(args.infile, "rb")

    # NOTE Not sure why but piping to rpmfile doesn't work unless we read
    # everything first into a buffer. Probably because of seek().
    buf = io.BytesIO(args.infile.read())

    output = {}

    if args.list:
        output["list"] = []
        with rpmfile.open(fileobj=buf) as rpm:
            for rpminfo in rpm.getmembers():
                print(rpminfo.name)
                output["list"].append(rpminfo.name.split("/"))
    elif args.info:
        output["info"] = ""
        with rpmfile.open(fileobj=buf) as rpm:
            headers_titles = {
                "name": "Name",
                "version": "Version",
                "release": "Release",
                "arch": "Architecture",
                "group": "Group",
                "size": "Size",
                "copyright": "License",
                "signature": "Signature",
                "sourcerpm": "Source RPM",
                "buildtime": "Build Date",
                "buildhost": "Build Host",
                "url": "URL",
                "summary": "Summary",
                "description": "Description",
            }
            for header in headers_titles:
                value = rpm.headers.get(header)
                if isinstance(value, bytes):
                    value = value.decode()
                if header == "buildtime":
                    value = datetime.fromtimestamp(value).strftime("%c")
                if header == "description":
                    value = "\n" + value
                line = "%s: %s" % (headers_titles.get(header).ljust(12), value)
                print(line)
                output["info"] += line + "\n"
    elif args.extract:
        output["extracted"] = []
        dest = os.path.abspath(args.dest)
        if not os.path.isdir(dest):
            raise FileNotFoundError(dest + " is not a directory")
        with rpmfile.open(fileobj=buf) as rpm:
            for rpminfo in rpm.getmembers():
                with rpm.extractfile(rpminfo.name) as rpmfileobj:
                    dirs = rpminfo.name.split("/")
                    filename = dirs.pop()
                    if dirs:
                        dirs_path = os.path.abspath(os.path.join(dest, *dirs))
                        if not dirs_path.startswith(dest):
                            raise ValueError("Attempted path traveral: " + dirs_path)
                        if not os.path.isdir(dirs_path):
                            os.makedirs(dirs_path)
                    target = os.path.abspath(os.path.join(dest, *(dirs + [filename])))
                    if not target.startswith(dest):
                        raise ValueError("Attempted path traveral: " + target)
                    outfile = open(target, "wb")
                    try:
                        outfile.write(rpmfileobj.read())
                    finally:
                        outfile.close()
                    if args.verbose:
                        print(target)
                    output["extracted"].append(rpminfo.name.split("/"))
    else:
        raise Exception("Nothing to do")

    buf.close()
    args.infile.close()

    return args, output
